use polldance::{net::TcpStream, PollManager, Readiness};
use std::sync::Arc;

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let mut poller = PollManager::new()?;

    let mut builder = Arc::new(TcpStream::connect([127, 0, 0, 1].into(), 4444)?);
    let mut builder_key = poller.register(Arc::clone(&builder), Readiness::write());

    let stream = 'l: loop {
        for (key, readiness) in poller.poll(None)? {
            if key == builder_key && readiness.is_write() {
                poller.deregister(builder_key);

                let bldr = match Arc::try_unwrap(builder) {
                    Ok(bldr) => bldr,
                    Err(_) => unreachable!("Only one Arc clone now"),
                };

                println!("Trying connect");
                match bldr.try_finish()? {
                    Ok(stream) => break 'l Arc::new(stream),
                    Err(bldr) => {
                        builder = Arc::new(bldr);

                        builder_key = poller.register(Arc::clone(&builder), Readiness::write());
                    }
                }
            }
        }
    };

    println!("Connected");

    let mut notify_token = poller.notifier();
    let handle = std::thread::spawn(move || {
        loop {
            std::thread::sleep(std::time::Duration::from_secs(5));
            match notify_token.notify() {
                Ok(c) => notify_token = c,
                Err(_) => break,
            }
        }
        println!("broken");
    });

    let stream_key = poller.register(Arc::clone(&stream), Readiness::read() | Readiness::hangup());
    let mut bytes = vec![];

    'l2: loop {
        for (key, readiness) in poller.poll(None)? {
            if key == stream_key {
                if readiness.is_read() {
                    loop {
                        let mut buf = [0; 1024];

                        match rustix::io::read(&*stream, &mut buf) {
                            Ok(n) => {
                                if n == 0 {
                                    break 'l2;
                                }
                                let should_break = n < buf.len();
                                bytes.extend_from_slice(&buf[0..n]);
                                if should_break {
                                    break;
                                }
                            }
                            Err(e)
                                if e == rustix::io::Error::WOULDBLOCK
                                    || e == rustix::io::Error::AGAIN =>
                            {
                                break;
                            }
                            Err(e) => return Err(e.into()),
                        }
                    }

                    poller.update_interests(&stream_key, Readiness::write() | Readiness::hangup());
                }
                if readiness.is_hangup() {
                    break 'l2;
                }
                if readiness.is_write() {
                    match rustix::io::write(&*stream, &bytes) {
                        Ok(n) if n == bytes.len() => {
                            bytes = Vec::new();
                            poller.update_interests(
                                &stream_key,
                                Readiness::read() | Readiness::hangup(),
                            );
                        }
                        Ok(n) => {
                            bytes = Vec::from(&bytes[n..]);
                        }
                        Err(e)
                            if e == rustix::io::Error::WOULDBLOCK
                                || e == rustix::io::Error::AGAIN =>
                        {
                            continue;
                        }
                        Err(e) => return Err(e.into()),
                    }
                }
            }
        }
        println!("Woken");
    }

    poller.deregister(stream_key);
    drop(poller);

    println!("Disconnected");

    handle.join().unwrap();

    Ok(())
}
